"""Module designed to handle all arguments"""
from argparse import ArgumentParser, Namespace
from re import match
from typing import Optional, Tuple
import os
import sys
import pkg_resources

from ansible_deployer.modules.validators.validate import Validators
from ansible_deployer.modules import globalvars


class CliInput:
    """Class designed to parse and validate command line arguments"""

    @staticmethod
    def create_parser() -> ArgumentParser:
        """Generic function to parse options for all commands"""
        parser = ArgumentParser(add_help=True)

        parser.add_argument("subcommand", nargs='*', default=None, metavar="SUBCOMMAND",
                            help='Specify subcommand to execute. Available commands: ' +
                                 str(globalvars.SUBCOMMANDS))
        parser.add_argument("--check-mode", "-C", action="store_true",
                            help='Enable --check-mode/-C when using default runner'
                                 ' (ansible-playbook).')
        parser.add_argument("--commit", "-c", nargs=1, default=[None], metavar="COMMIT",
                            help='Provide commit ID.')
        parser.add_argument("--conf-dir", nargs=1, default=[None], metavar="conf_dir",
                            help='Use non-default configuration directory, only allowed for'
                                 ' non-binarized exec')
        parser.add_argument("--conf-validation", default=False, action='store_true',
                            help='Execute configuration files validation and exit program.')
        parser.add_argument("--debug", "-d", default=False, action="store_true",
                            help='Print debug output. This option does not affect runner output'
                                 ' (use --raw-runner-output for that).')
        parser.add_argument("--dry-mode", "-D", action="store_true",
                            help='Execute default runner (ansible-playbook) with'
                                 ' "ansible_deployer_dry_mode" tag, triggering only required'
                                 ' variable validation in pre_tasks. This tag is not predefined!')
        parser.add_argument("--infrastructure", "-i", nargs=1, default=[None],
                            metavar="INFRASTRUCTURE",
                            help='Specify infrastructure for deploy.')
        parser.add_argument("--inventory", "-I", nargs=1, default=[None], metavar="INVENTORY",
                            help='Specify inventory, only allowed if not set in configuration')
        parser.add_argument("--keep-locked", "-k", default=False, action='store_true',
                            help='Keep infrastructure locked after task execution.')
        parser.add_argument("--limit", "-l", nargs=1, default=[None], metavar="[LIMIT]",
                            help='Limit task execution to specified host.')
        parser.add_argument("--no-color", default=False, action="store_true",
                            help='Disable coloring of console messages.')
        parser.add_argument("--no-lock", default=False, action="store_true",
                            help='Do not lock the infrastructure')
        parser.add_argument("--raw-runner-output", default=False, action="store_true",
                            help='Print original messages in real time during runner execution.')
        parser.add_argument("--runner-options", nargs=1, default=[None],
                            metavar="'\"--a -b --opt2\"'",
                            help='Quote-enclosed raw string of flags to be passed.')
        parser.add_argument("--runner-plugins", nargs=1, default=[None], metavar='TASK_NAME',
                            help='Provide comma-separated list of plugins to enable.')
        parser.add_argument("--runner-raw-file", default=False, action="store_true",
                            help='Print original messages to main log file.')
        parser.add_argument("--runner-stdout", nargs=1, default=[None], metavar='STDOUT_PLUGIN',
                            help=
                            'Provide name of runner stdout callback plugin you would like to use.')
        parser.add_argument("--runner-verbosity", nargs=1, default=[1], metavar='N',
                            help='Set runner verbosity to N*v.')
        parser.add_argument("--self-setup", nargs=1, default=[None], metavar="LOCAL_SETUP_PATH",
                            help='Setup repo outside of workdir in requested path. This option'
                                 ' applies only to infrastructures with allow_user_checkout enabled'
                                 ' in infra config!')
        parser.add_argument("--stage", "-s", nargs=1, default=[None], metavar="STAGE",
                            help='Specify stage type. Available types are: "testing" and "prod".')
        parser.add_argument("--syslog", default=False, action="store_true", help='Log warnings and'
                            ' errors to syslog. --debug doesn\'t affect this option!')
        parser.add_argument("--task", "-t", nargs=1, default=[None], metavar='TASK_NAME',
                            help='Provide task_name.')
        parser.add_argument("--version", "-v", default=False, action="store_true",
                            help='Display app version and exit.')

        return parser

    def parse_arguments(self) -> dict:
        """
        Generic function that returns options dict after triggering command line arguments parsing
        and validation.
        """
        return self.validate_arguments(self.create_parser().parse_args())

    def validate_arguments(self, arguments: Namespace) -> dict:
        """
        Generic function that returns options dict after triggering command line arguments
        validation
        """
        return self.validate_rest_arguments(arguments, *self.validate_init_arguments(arguments))

    @staticmethod
    def validate_init_arguments(arguments: Namespace) -> Tuple[str, str]:
        """Validate selected parsed arguments for specific handling"""
        if arguments.version:
            version = pkg_resources.require("ansible_deployer")[0].version
            print(f"ansible-deployer version: {version}")
            sys.exit(0)

        if arguments.no_color:
            PRINT_FAIL = PRINT_END = ""
        else:
            PRINT_FAIL = globalvars.PRINT_FAIL
            PRINT_END = globalvars.PRINT_END

        if not arguments.subcommand:
            sub_string = ", ".join(globalvars.SUBCOMMANDS).strip(", ")
            print(f"{PRINT_FAIL}[CRITICAL]: First positional argument (subcommand) is required!"
                  f" Available commands are: {sub_string}.{PRINT_END}")
            sys.exit(57)

        return PRINT_END, PRINT_FAIL

    def validate_rest_arguments(self, arguments: Namespace, print_end: str, print_fail: str
                                ) -> dict:
        """Validate remaining parsed arguments and collect them into dictionary"""
        options = {}
        options["subcommand"] = arguments.subcommand[0].lower()
        Validators.verify_subcommand(options["subcommand"], arguments.no_color)
        Validators.verify_switches(arguments.subcommand, arguments.no_color)

        options["switches"] = arguments.subcommand[1:]
        options["check_mode"] = arguments.check_mode
        options["commit"] = arguments.commit[0]
        options["conf_dir"] = os.path.abspath(arguments.conf_dir[0]) if arguments.conf_dir[0]\
            else None
        options["conf_val"] = arguments.conf_validation
        options["debug"] = arguments.debug
        options["dry_mode"] = arguments.dry_mode
        options["infra"] = arguments.infrastructure[0]
        options["inventory"] = arguments.inventory[0]
        options["keep_locked"] = arguments.keep_locked
        options["limit"] = arguments.limit[0]
        options["lock"] = not arguments.no_lock
        options["no_color"] = arguments.no_color
        options["raw_output"] = arguments.raw_runner_output
        options["runner_opts"] = self.validate_option_string(arguments.runner_options[0],
                                                             print_fail, print_end)
        options["runner_plugins"] = self.validate_ansible_callback(
            arguments.runner_plugins[0], print_fail, print_end)
        options["runner_raw_file"] = arguments.runner_raw_file
        options["runner_stdout"] = self.validate_ansible_stdout_callback(arguments.runner_stdout,
                                                                         print_fail, print_end)
        options["runner_verb"] = self.validate_integer(arguments.runner_verbosity[0], print_fail,
                                                       print_end)
        options["self_setup"] = os.path.abspath(arguments.self_setup[0]) if arguments.self_setup[0]\
            else None
        options["stage"] = arguments.stage[0]
        options["syslog"] = arguments.syslog
        options["task"] = arguments.task[0]

        return options

    @staticmethod
    def validate_ansible_stdout_callback(stdout_plugin: str, print_fail: str, print_end: str
                                         ) -> Optional[str]:
        """Validate whether plugin (specified via --runner-stdout option) is supported"""
        try:
            lplugin = stdout_plugin[0].lower()
            if lplugin in globalvars.SUPPORTED_STDOUT_CALLBACK_PLUGINS:
                return lplugin

            print(f"{print_fail}[CRITICAL]: Unsupported runner stdout callback plugin!"
                  f"{print_end}")
            sys.exit(57)
        except AttributeError:
            return None

    @staticmethod
    def validate_ansible_callback(stdout_plugin: str, print_fail: str, print_end: str
                                  ) -> Optional[list]:
        """Validate whether plugin (specified via --runner-plugin option) is supported"""
        if stdout_plugin:
            try:
                plugins = stdout_plugin.split(",")
                if all(plugin in globalvars.SUPPORTED_CALLBACK_PLUGINS for plugin in plugins):
                    return plugins

                print(f"{print_fail}[CRITICAL]: Unsupported runner callback plugin!"
                      f"{print_end}")
                sys.exit(57)
            except Exception:
                print(f"{print_fail}[CRITICAL]: Invalid format of runner callback plugin!"
                      f"{print_end}")
                sys.exit(57)
        else:
            return []

    @staticmethod
    def validate_integer(number: str, print_fail: str, print_end: str) -> Optional[int]:
        """Validate if number is an integer"""
        try:
            return int(number)
        except ValueError:
            print(f"{print_fail}[CRITICAL]: Not an integer!{print_end}")
            sys.exit(57)

    @staticmethod
    def validate_option_string(options: str, print_fail: str, print_end: str) -> Optional[str]:
        """Validate if provided --runner-options are strings matching certain pattern """
        if options:
            if isinstance(options, str) and match(r"[ -A-za-z0-9]+", options):
                return options

            print(f"{print_fail}[CRITICAL]: Invalid format of runner options!{print_end}")
            sys.exit(57)
        else:
            return None



"""Module for spawning customized loggers based on generic AnsibleDeployerLogger"""
from os.path import join as ospjoin
from typing import Tuple

from . import AnsibleDeployerLogger


class AppLogger(AnsibleDeployerLogger):
    """
    Create main application logger and add it's handlers
    memory handler exists until file handler is added in main function
    syslog handler will be added only if --syslog flag was set by operator
    """

    def __init__(self, options: dict):
        super().__init__("ansible-deployer_app_logger", options)
        self.add_syslog_handler()
        self.add_memory_handler()
        self.add_console_handler(self.console_formatter)


class RunLogger(AnsibleDeployerLogger):
    """
    Logger object created for each playitem from
    ansible_deployer.modules.runners.run.Runners.run_playitem()
    """

    def __init__(self, options: dict, workdir: str, playitem_name: str, inventory_name: str):
        name, log_path = self.set_runner_vars(workdir, playitem_name, inventory_name)
        super().__init__(name, options)
        self.log_path = log_path
        self.add_syslog_handler()

    def set_runner_vars(self, workdir: str, playitem: str, inventory: str) -> Tuple[str, str]:
        """Set logger_name and log_path for current playitem in sequence"""
        playitem_fmt = playitem.replace(" ", "_")
        logger_name = f"ansible-deployer_runner_{playitem_fmt}_{inventory}_logger"
        log_path = ospjoin(workdir, f"rawlog_{playitem_fmt}_{inventory}.log")

        return logger_name, log_path

    def add_raw_handlers(self, main_log_path: str) -> None:
        """Add handlers to runner logger triggered by --runner-raw-* options"""
        if self.options["raw_output"]:
            self.add_console_handler(self.raw_formatter)

        if self.options["runner_raw_file"]:
            self.add_file_handler(main_log_path, self.raw_formatter)
        else:
            self.add_file_handler(self.log_path, self.raw_formatter)



"""Module for handling of logging and log files"""

import logging
from logging import handlers as log_han


class CustomFormatter(logging.Formatter):
    """Class adding colours to console logger"""

    def __init__(self, formatter):
        super().__init__()
        grey = "\x1b[0;38m"
        light_green = "\x1b[1;32m"
        yellow = "\x1b[0;33m"
        red = "\x1b[0;31m"
        light_red = "\x1b[1;31m"
        reset = "\x1b[0m"

        self.FORMATS = {
            logging.DEBUG: light_green + formatter + reset,
            logging.INFO: grey + formatter + reset,
            logging.WARNING: yellow + formatter + reset,
            logging.ERROR: red + formatter + reset,
            logging.CRITICAL: light_red + formatter + reset
        }

    def format(self, record):
        """Colour logged message depending on log level"""
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class AnsibleDeployerLogger:
    """Class handling creating logger and logging handlers"""

    formatter = logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s")
    basic_formatter = "%(asctime)s [%(levelname)s]: %(message)s"
    console_formatter = "\n%(asctime)s [%(levelname)s]: %(message)s\n"
    raw_formatter = "%(message)s"

    def __init__(self, name: str, options: dict):
        self.name = name
        self.options = options
        self.logger = self.create_logger()

    def create_logger(self) -> logging.Logger:
        """Function to create logging objects"""
        logger = logging.getLogger(self.name)
        logger.setLevel(logging.DEBUG)

        return logger

    def add_syslog_handler(self):
        """Add syslog handler sending messages through /dev/log"""
        if self.options["syslog"]:
            rsys_handler = log_han.SysLogHandler(address="/dev/log")
            rsys_handler.setFormatter(self.formatter)
            rsys_handler.setLevel(logging.WARNING)
            self.logger.addHandler(rsys_handler)

    def add_memory_handler(self):
        """Add memory handler with hardcoded limit of 15 messages"""
        memory_handler = log_han.MemoryHandler(15, flushLevel=logging.DEBUG, flushOnClose=False)
        memory_handler.setFormatter(self.formatter)
        memory_handler.setLevel(logging.DEBUG)
        self.logger.addHandler(memory_handler)

    def add_console_handler(self, console_formatter_type: str):
        """Add console handler influenced by operator arguments"""
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(logging.Formatter(console_formatter_type)
                                     if self.options["no_color"]
                                     else CustomFormatter(console_formatter_type))
        console_handler.setLevel(logging.DEBUG if self.options["debug"] else logging.INFO)
        self.logger.addHandler(console_handler)

    def add_file_handler(self, log_path: str, file_formatter_type: str):
        """Function adding file handler to existing logger"""
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(logging.Formatter(file_formatter_type))
        file_handler.setLevel(logging.DEBUG)
        self.logger.addHandler(file_handler)

    def flush_memory_handler(self, subcommand_flag: bool, syslog: bool):
        """Flush initial log messages from memory handler to logfile"""
        if syslog:
            # handler[1] - MemoryHandler, handler[3] - FileHandler
            if subcommand_flag:
                self.logger.handlers[1].setTarget(self.logger.handlers[3])
                self.logger.handlers[1].flush()
            self.logger.handlers[1].close()
        else:
            # handler[0] - MemoryHandler, handler[2] - FileHandler
            if subcommand_flag:
                self.logger.handlers[0].setTarget(self.logger.handlers[2])
                self.logger.handlers[0].flush()
            self.logger.handlers[0].close()



"""Main module for ansible-deploy"""

import os
import sys
import datetime
import errno
from ansible_deployer.modules.configs.config import Config
from ansible_deployer.modules.locking.locking import Locking
from ansible_deployer.modules.misc import utils as mutils
from ansible_deployer.modules.misc.arguments import CliInput
from ansible_deployer.modules.outputs import blocks
from ansible_deployer.modules.outputs.loggers import AppLogger
from ansible_deployer.modules.validators.validate import Validators
from ansible_deployer.modules.runners.run import Runners
from ansible_deployer.modules.database.creator import DbSetup
from ansible_deployer.modules.database.writer import DbWriter
from ansible_deployer.modules import globalvars


def main(options: dict):
    """ansible-deployer main function"""
    start_ts_raw = datetime.datetime.now()
    start_ts = start_ts_raw.strftime("%Y%m%d_%H%M%S")

    if len(sys.argv) < 2:
        print(f"{globalvars.PRINT_FAIL}[CRITICAL]: Too few arguments{globalvars.PRINT_END}",
              file=sys.stderr)
        sys.exit(2)

    logger = AppLogger(options)

    configuration = Config(logger.logger, options["conf_dir"])
    config = configuration.load_configuration()

    if options["subcommand"] in ("run", "verify"):
        workdir = mutils.create_workdir(start_ts, configuration.conf, logger.logger)
        log_path = os.path.join(
            workdir, configuration.conf["file_naming"]["log_file_name_frmt"].format(start_ts))
        logger.add_file_handler(log_path, logger.basic_formatter)
        logger.flush_memory_handler(True, options["syslog"])
    else:
        logger.flush_memory_handler(False, options["syslog"])

    validators = Validators(logger.logger)
    validators.validate_options(options)
    selected_items = validators.validate_option_values_against_config(config, options)

    if options["subcommand"] in ("run", "verify"):
        os.chdir(options["self_setup"] if options["self_setup"] else workdir)

    user_groups = mutils.get_all_user_groups(logger.logger)

    if options["conf_val"]:
        logger.logger.info("Validation of configuration files was successful, program will exit now"
                           ".")
        sys.exit(0)

    if options["subcommand"] == "show":
        mutils.show_deployer(config, options)
    else:
        inv_file = mutils.get_inventory_file(config, options, logger.logger)

        lockpath = os.path.join(os.path.join(configuration.conf["global_paths"]["work_dir"],
                                "locks") , inv_file.lstrip(f".{os.sep}").replace(os.sep, "_"))
        lock = Locking(logger.logger, options["keep_locked"], (options["infra"], options["stage"]),
                       configuration.conf)
        if options["subcommand"] in ("run", "verify"):
            if not validators.verify_task_permissions(selected_items, user_groups, config):
                logger.logger.critical("Task forbidden")
                sys.exit(errno.EPERM)
            db_connector, db_path = DbSetup.connect_to_db(DbSetup(logger.logger, start_ts,
                                                          configuration.conf, options))
            db_writer = DbWriter(logger.logger, db_connector, db_path)
            runner = Runners(logger.logger, lock, workdir, start_ts_raw,
                             config["tasks"]["setup_hooks"], log_path, db_path)
            if not options["self_setup"]:
                runner.setup_ansible(selected_items["commit"], configuration.conf_dir)
            if options["lock"]:
                lock.lock_inventory(lockpath)
            sequence_record_dict = runner.run_playitem(configuration.conf["callback_settings"],
                                                       config, options, inv_file, lockpath,
                                                       db_writer)
            if options["lock"]:
                lock.unlock_inventory(lockpath)
            blocks.log_exit_messages(logger.logger, log_path, db_path)
            db_writer.finalize_db_write(sequence_record_dict, False)
        elif options["subcommand"] == "lock":
            lock.lock_inventory(lockpath)
        elif options["subcommand"] == "unlock":
            lock.unlock_inventory(lockpath)

    sys.exit(0)


def ansible_deployer():
    """ansible-deployer endpoint function"""
    return main(CliInput.parse_arguments(CliInput()))


if __name__ == "__main__":
    ansible_deployer()



"""
This module's intention is to aggregate larger (>3) or inconvenient-to-be-placed-in-code logging or
print statements into functions
"""
from logging import Logger
from os import sep as ossep


def log_exit_messages(logger: Logger, log_path: str, db_path: str) -> None:
    """Log info about logfile and db at the end of exec"""
    exe_id = log_path.rsplit(ossep, 2)[1]
    logger.info(f"This execution received ID: {exe_id}")
    logger.info("More information on this execution can be acquired in:")
    logger.info(f"logfile: {log_path}")
    logger.info(f"database: {db_path}")
